
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
# import faiss
from subgrad import *
from sklearn.ensemble import RandomForestRegressor
from sklearn.calibration import CalibratedClassifierCV
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.model_selection import KFold
from sklearn.model_selection import StratifiedKFold
from joblib import Parallel, delayed
from copy import deepcopy
import pickle

class config():
  # metadata = {'render.modes': ['human']}

  def __init__(self, horizon, p, K, J, M, S0, action_price_map, XGenerator, arrival_type_dens, Y_AX, gamma = 1, nA = 2, seed =1):
    # Take in functional parametrization of X, and conditional response Y| A,X
    # M resource consumption matrix
    # product arrival rate Z
    # initial capacity S0
    # S0 is list of integer starting capacities for each product
    # dimension init parameters
    # XGenerator(t,k)
    # arrival_type_dens(t,k)
    # Y_AX(price, x,k)


    self.horizon = horizon
    self.p = p # context dimension
    self.J = J # number of resources
    self.K = K # number of products
    self.M = M
    self.XGenerator = XGenerator
    self.arrival_type_dens = arrival_type_dens
    self.Y_AX = Y_AX
    self.nA = nA
    self.action_price_map = action_price_map #k x nA map of, for each product, prices for each action
    self.pS = len(S0)
    self.gamma = gamma
    return



class nrm_config(config):
  def __init__(self, horizon, p, K, J, M, S0, action_price_map, XGenerator, arrival_type_dens, Y_AX, nA = 2, seed =1):
    super().__init__(horizon, p, K, J, M, S0, action_price_map, XGenerator, arrival_type_dens, Y_AX, nA = 2, seed =1)

  # these assume inputs in "marginal S" space
  def R_nrm(self,s,a,sprime):
    Y=(sprime - s) # 0 or 1; purchase/nopurchase
    if Y<0:
      price =self.action_price_map[a]
      return price
    else: return 0

  def R_nrm_multiarrival(self, s, a, sprime,k):
    sprime=np.array(sprime); s=np.array(s)
    Y = sum(sprime - s)  # 0 or 1; purchase/nopurchase
    diff = s - sprime
    # which object sold
    if Y < 0:
      match = np.asarray([ (diff == row).all() for row in self.M ])
      if sum(match.astype(int)) > 0:
        k_ = np.where([ (diff == row).all() for row in self.M ])[0].item()
        if k==k_:
          price = self.action_price_map[k_, a]
          return price

        else: # product type doesn't match
          return 0
    else:
      return 0


  def nrm_mapper(self, s, sprime, y):
    # input states as tuples
    # return 0 if infeasible; 1 if feasible
    diff = s-sprime  # 0 or 1; purchase/nopurchase
    # which object sold
    if sprime<0:
      return 0
    if y==0:
      if (diff)==0: # didn't sell
        return 1
      else: return 0
    else: # sold
      if diff==1:
        return 1
      else: return 0

  def nrm_mapper_multiarrival(self, s, sprime, y, k):
    # input states as tuples
    s = np.asarray(s); sprime = np.asarray(sprime)
    # return 0 if infeasible; 1 if feasible
    diff = s-sprime  # 0 or 1; purchase/nopurchase

    if (sprime<0).any():
      return 0
    if (sprime>s).any():
      return 0

    if y==0:
      # only valid if didn't sell
      if sum(np.abs(diff))==0:
        return 1
      else:
        return 0
    else: # sold
      # check state difference matches resource allocation matrix
      match = np.asarray([ (diff == row).all() for row in self.M ])
      if sum(match.astype(int)) > 0:
        k_ = np.where([ (diff == row).all() for row in self.M ])[0].item()
        if k_ == k: # product transition matches given product arrival type
          return 1
        else: return 0
      else: return 0



class FaissKNeighbors:
    def __init__(self, k=5):
        self.index = None
        self.y = None
        self.k = k

    def fit(self, X, y):
        self.index = faiss.IndexFlatL2(X.shape[1])
        self.index.add(X.astype(np.float32))
        self.y = y

    def predict(self, X):
        distances, indices = self.index.search(X.astype(np.float32), k=self.k)
        votes = self.y[indices]
        predictions = np.array([np.argmax(np.bincount(x)) for x in votes])
        return predictions

def sample_trajectory(env, policy, config):
  # reset env
  env.reset()
  horizon = config.horizon
  J = config.J
  p = config.p
  S_ = np.zeros([horizon, config.pS])# structural state
  X_ = np.zeros([horizon, p]); K_ = np.zeros(horizon); R_ = np.zeros(horizon); Y_ = np.zeros(horizon)
  X_[0, :] = env.state[J:J + p]
  K_[0] = env.state[-1]
  S_[0,:] = env.state[:J]
  A_ = np.zeros(horizon)
  for t in range(horizon):
    # policy takes in X,k,S
    action = policy(X_[t,:], K_[t], S_[0:J], t)

    [S, R, epover, null] = env.step(action)

    A_[t] = action
    R_[t] = R
    Y_[t] = R > 0
    # Step returns x_t+1
    if t < horizon-1:
      S_[t + 1, :] = S[0:J]
      X_[t + 1, :] = S[J:J + p]
      K_[t + 1] = S[-1]

  return [S_, X_, A_, K_, R_, Y_]


class singleClassPredictor():
  def __init__(self,n_classes):
    self.n_classes = n_classes
  def fit(self, X, Y):
    y = np.unique(Y)
    self.y = int(y)

  def predict(self, x):
    return self.y

  def predict_proba(self, x):
    if len(x.shape)>1:
      proba = np.zeros([x.shape[0],self.n_classes])
      proba[:,self.y] = 1
    else:
      proba = np.zeros(self.n_classes)
      proba[self.y] = 1
    return proba


def fit_LR_for_XY(X, Y, regressor=LogisticRegression, test_size=0.2,verbose=False, n_classes=2, calibrate=False):
  # X_train, X_test, y_train, y_test = train_test_split(
  #   X, Y, test_size=test_size)  # , random_state=42)
  X_train=X;y_train=Y
  if len(np.unique(y_train))>1: # if enough classes appear
    LR = regressor()
    # if calibrate and X_train.shape[0]>500:
    #   if len(np.unique(y_train))>1:
    #     try:
    #       LR=CalibratedClassifierCV(base_estimator=LR) # CV will fail if folds don't contain enough classes: if this happens, don't calibrate
    #     except ValueError as e:
    #       LR = regressor()
    LR.fit(X_train, y_train)
  else: # finite-sample issue with only instances of one class appearing
    print("single-class predictor")
    LR = singleClassPredictor(n_classes)
    LR.fit(X_train,y_train)

  return LR


# train for each product arrival type
def get_Y_AX_actionspace_multinomial(config_, X__, A__, Y__, K__):
  # input: config; data X__, Y__, K__ (over trajectories, N x horizon x p)
  # output: multinomial/multiproduct Y mapper, Y | A,X models
  # learns a nuisance model for every action a
  [N, horizon, p] = X__.shape
  Y_AXs = dict()

  for k in range(config_.K):
    for a in range(config_.nA):
      k_inds = np.where((K__.reshape(N * horizon) == k) & (A__.reshape(N * horizon) == a) )[0]
      Y_AX_LR = fit_LR_for_XY(X__.reshape([N * horizon, p])[k_inds, :], Y__.reshape([N * horizon])[k_inds])
      Y_AXs[tuple([k,a])] = Y_AX_LR

  def Y_AX_actionspace(y, a, x, k):
    # assume Y model given as dictionary indexed by k, arrival type
    k = int(k); a = int(a); x=x.reshape(1, -1)
    ypred = Y_AXs[tuple([k,a])].predict_proba(x).flatten()  # self.classes_ ordering
    ypred=ypred[y]
    return ypred

  return [Y_AX_actionspace, Y_AXs]


def get_e_AX_multinomial(config_, X__, A__, K__):
  [N, horizon, p] = X__.shape
  e_AXs = dict()
  for k in range(config_.K):
    k_inds = np.where( (K__.reshape(N * horizon) == k) )[0]
    A_X_LR = fit_LR_for_XY(X__.reshape([N * horizon, p])[k_inds], A__.reshape([N * horizon])[k_inds])
    e_AXs[k] = A_X_LR
  def e_AX(a, x, k):
    x=x.reshape(1, -1)
    return e_AXs[k].predict_proba(x).flatten()[a]  # return prob A=a|X
  return [e_AX, e_AXs ]

def get_Y_AX_singlearrival(config_, X__, A__, Y__,calibrate=False,regressor=LogisticRegression):
  # input: config; data X__, Y__, K__ (over trajectories, N x horizon x p)
  # output: multinomial/multiproduct Y mapper, Y | A,X models
  # learns a nuisance model for every action a


  Y_AXs = dict()
  for a in range(config_.nA):
    if X__.ndim > 2:  # if trajectory-time structured
      [N, horizon, p] = X__.shape
      a_inds = np.where((A__.reshape(N * horizon) == a) )[0]
      Y_AX_LR = fit_LR_for_XY(X__.reshape([N * horizon, p])[a_inds], Y__.reshape([N * horizon])[a_inds],regressor=regressor,calibrate=calibrate)
    else:
      [N, p] = X__.shape
      a_inds = np.where((A__ == a))[0]
      Y_AX_LR = fit_LR_for_XY(X__[a_inds], Y__[a_inds],
                              regressor=regressor, calibrate=calibrate)
    Y_AXs[a] = Y_AX_LR
  def Y_AX_actionspace(y, a, x):
    # reshape if passed matrix X vs vector X
    if len(x.shape) ==1:
      xvec=True
      x = x.reshape(1, -1)
    else: xvec=False
    # assume Y model given as dictionary indexed by k, arrival type
    a = int(a);

      # self.classes_ ordering
    if not xvec:
      ypred = Y_AXs[a].predict_proba(x)
      ypred=ypred[:,y]
    else:
      ypred = Y_AXs[a].predict_proba(x).flatten()
      ypred=ypred[y]
    return ypred

  return [Y_AX_actionspace, Y_AXs]


def get_e_AX_singlearrival(config_, X__, A__):
  if X__.ndim > 2: # if trajectory-time structured
    [N, horizon, p] = X__.shape
    A_X_LR = fit_LR_for_XY(X__.reshape([N * horizon, p]), A__.reshape([N * horizon]))
  else:
    A_X_LR = fit_LR_for_XY(X__,A__)

  def e_AX(a, x):
    if x.ndim>1:
      N=x.shape[0]
      if hasattr(a, "__len__"):
        return np.asarray([A_X_LR.predict_proba(x)[i, a[i]] for i in range(N)]).flatten()
      else:
        return A_X_LR.predict_proba(x)[:,a]
    else:
      x=x.reshape(1, -1)
      return A_X_LR.predict_proba(x).flatten()[a]  # return prob A=a|X
  return [e_AX, A_X_LR ]


# learn outcome model
def IPW_score(X,A,a,Y,y,e_AX,Y_AX):
  return (Y==y)*(A==a)*1.0/e_AX(a,X)
def DM_score(X,A,a,Y,y,e_AX,Y_AX):
  return Y_AX(y,a,X)
def DR_score(X,A,a,Y,y,e_AX,Y_AX):
  return (A == a) * 1.0 / e_AX(a, X)*(Y == y)-Y_AX(y,a,X) + Y_AX(y,a,X)

def IPW_score_multiarrival(X,A,a,Y,y,e_AX,Y_AX,K,k):
  return (K==k)*(Y==y)*(A==a)*1.0/e_AX(a,X,k)
def DM_score_multiarrival(X,A,a,Y,y,e_AX,Y_AX,K,k):
  score = Y_AX(y,a,X,k)
  return score
def DR_score_multiarrival(X,A,a,Y,y,e_AX,Y_AX,K,k):
  score = (K==k)*(A==a) * 1.0 / e_AX(a, X, k)*((Y==y)-Y_AX(y,a,X,k)) + Y_AX(y,a,X,k)
  return score



def IPW_score_multiarrival_vec(X,A,a,Y,y,e_AX,Y_AX,K,k, mult=None):
  props = np.asarray(e_AX(a,x,k) for x in X)
  if mult is not None:
    return mult*(K==k)*(Y==y)*(A==a)*1.0/props
  else:
    return (K==k)*(Y==y)*(A==a)*1.0/props

def DM_score_multiarrival_vec(X,A,a,Y,y,e_AX,Y_AX,K,k, mult=None):
  n = X.shape[0]
  scores = np.array([Y_AX(y,a,X[i,:],k)  for i in range(n)])
  return scores
def DR_score_multiarrival_vec(X,A,a,Y,y,e_AX,Y_AX,K,k, mult=None):
  n = X.shape[0]
  scores = np.array([Y_AX(y,a,X[i,:],k)  for i in range(n)])
  props = np.asarray(e_AX(a, x, k) for x in X)
  if mult is not None:
    final = (((K==k)*(A==a) * (1.0 / props)*((Y==y)-scores)) + scores)*mult
  else:
    final = ((K==k)*(A==a) * (1.0 / props)*((Y==y)-scores)) + scores
  return final

def IPW_score_singlearrival_vec(X,A,a,Y,y,e_AX,Y_AX, mult=None):
  props = np.asarray(e_AX(a,x,k) for x in X)
  if mult is not None:
    return mult*(Y==y)*(A==a)*1.0/props
  else:
    return (Y==y)*(A==a)*1.0/props

def DM_score_singlearrival_vec_a(X,A,a,Y,y,e_AX,Y_AX,mult=None):
  n = X.shape[0]
  if mult is not None:
    scores = Y_AX(y,a,X) * mult
    # scores = np.array([Y_AX(y,a,X[i,:])  for i in range(n)]) * mult
  else:
    scores = Y_AX(y,a,X)
    # scores = np.array([Y_AX(y, a, X[i, :]) for i in range(n)])
  return scores

def DR_score_singlearrival_vec(X,A,a,Y,y,e_AX,Y_AX, mult=None):
  n = X.shape[0]
  scores = np.array([Y_AX(y,a,X[i,:])  for i in range(n)])
  props = np.asarray([e_AX(int(A[i]), X[i,:]) for i in range(n) ])
  if mult is not None:
    final = (((A==a) * (1.0 / props)*((Y==y)-scores)) + scores)*mult
  else:
    final = ((A==a) * (1.0 / props)*((Y==y)-scores)) + scores
  return final

rd = np.array([[ 0.65953346,  0.43524589,  0.70816539, -0.44480714, -0.14945422,
      -2.45490417]])


def R_AX(price, x, prob=True):
  X_ = np.hstack([x, price])

  if prob:
    return price * 5 + price * x.sum() * 10
  #         return prob*50
  else:
    Y = np.random.binomial(1, prob).item()
  return Y

def get_opt_V(config_, nS, mu, sigma, Y_AX):
  '''
  Get optimal policy/optimal value function for singlearrival dynamic pricing
  Optimize pointwise for X
  '''
  horizon = config_.horizon
  V = np.zeros([config_.horizon, nS])
  pct_A1 = np.zeros([config_.horizon, nS])
  purchase_prob = np.zeros([config_.horizon, nS])
  nX = 10000
  price_action_map = config_.action_price_map
  X = np.random.multivariate_normal(mu, sigma, size=nX)
  Y_1X = [Y_AX(price_action_map[1], X[i, :], 0) for i in range(nX)]
  Y_0X = [Y_AX(price_action_map[0], X[i, :], 0) for i in range(nX)]


  for t in reversed(range(horizon)):
    for s in range(nS):
      if t == horizon - 1:
        V[t, s] = 0
      else:
        # manual policy opt over fixed X distribution
        if s >= 1:

          # val1 = [Y_1X[i] * (R_AX(price_action_map[1], X[i, :]) + V[t + 1, s - 1]) + (1 - Y_1X[i]) * V[t + 1, s] for i in range(nX)]
          # val0 = [Y_0X[i] * (R_AX(price_action_map[0], X[i, :]) + V[t + 1, s - 1]) + (1 - Y_0X[i]) * V[t + 1, s] for i in range(nX)]
          val1 = [Y_1X[i] * (price_action_map[1] + V[t + 1, s - 1]) + (1 - Y_1X[i]) * V[t + 1, s] for i in range(nX)]
          val0 = [Y_0X[i] * (price_action_map[0] + V[t + 1, s - 1]) + (1 - Y_0X[i]) * V[t + 1, s] for i in range(nX)]
          opt_pol = [1 if val1[i] > val0[i] else 0 for i in range(nX)]
          pol_val = np.mean( [val1[i] if val1[i] > val0[i] else val0[i] for i in range(nX) ] )

          purchase_prob[t, s] = np.mean([Y_1X[i] if opt_pol[i] == 1 else Y_0X[i] for i in range(nX)])
          pct_A1[t, s] = np.mean(opt_pol)
          V[t, s] = pol_val
        else:
          V[t, s] = 0
  return [V, pct_A1, purchase_prob]

def expit(beta, x):
  return 1.0 / (1.0 + np.exp(-1 * np.dot(beta, x)))

def get_opt_V_constant(config_, nS, mu, sigma, Y_AX):
  '''
  Get optimal policy/optimal value function for singlearrival dynamic pricing
  constant actions
  '''
  horizon = config_.horizon
  V = np.zeros([config_.horizon, nS])
  nX = 10000
  price_action_map = config_.action_price_map
  X = np.random.multivariate_normal(mu, sigma, size=nX)


  Y_1X = [Y_AX(price_action_map[1], X[i, :], 0) for i in range(nX)]
  Y_0X = [Y_AX(price_action_map[0], X[i, :], 0) for i in range(nX)]
  for t in reversed(range(horizon)):
    # print(t)
    for s in range(nS):
      if t == horizon - 1:
        V[t, s] = 0
      else:# manual policy opt over fixed X distribution
        if s >= 1:
          # y1 = np.mean([Y_1X[i] * (R_AX(price_action_map[1],X[i,:]) + V[t + 1, s - 1]) + (1 - Y_1X[i]) * V[t + 1, s] for i in range(nX)])
          # y0 = np.mean([Y_0X[i] * (R_AX(price_action_map[0],X[i,:]) + V[t + 1, s - 1]) + (1 - Y_0X[i]) * V[t + 1, s] for i in range(nX)])
          y1 = np.mean([Y_1X[i] * (price_action_map[1] + V[t + 1, s - 1]) + (1 - Y_1X[i]) * V[t + 1, s] for i in range(nX)])
          y0 = np.mean([Y_0X[i] * (price_action_map[0] + V[t + 1, s - 1]) + (1 - Y_0X[i]) * V[t + 1, s] for i in range(nX)])
          V[t, s] = y1 if y1 > y0 else y0
        else:
          V[t, s] = 0
  return V



# def backwards_induction_with_estimates_single_arrival(config_, margSs,  # problem parameters
#                                         X_, A_, Y_,
#                                         G_s, R,  # problem mapping
#                                         score, e_AX, Y_AX):  # estimation strategy and nuisances
#   '''
#   Solve for policy by backwards induction
# 
#   '''
# 
#   dS = len(margSs)  # one dimensional representation
#   nA = config_.nA
#   [nX, horizon, p] = X_.shape
#   price_action_map = config_.action_price_map
#   V = np.zeros([horizon,dS])
#   for t in reversed(range(horizon)):
#     for s in margSs:
#       if t < horizon - 1:
#         if s >= 1:
#           scores = np.zeros([N, nA])
#           for a in range(nA): 
#             y=1 # est P(Y(a) = 1|X)
#             scores[:, a] = score(X_[:, t, :], A_[:, t], a, Y_[:, t], y, e_AX, Y_AX)
#               # range over x
#             y1 = np.mean(
#               [scores[i,1] * (price_action_map[1] + V[t + 1, s - 1]) + (1 - scores[i,1]) * V[t + 1, s] for i in range(nX)])
#             y0 = np.mean(
#               [scores[i,0] * (price_action_map[0] + V[t + 1, s - 1]) + (1 - scores[i,0]) * V[t + 1, s] for i in range(nX)])
#             V[t, s] = y1 if y1 > y0 else y0
# 
#             best_a = np.argmax(scores, axis=1)
#             V[t,s] += np.mean(scores[:, best_a])
#   return V
def DR_score_singlearrival_vec_a(X,A,a,Y,y,e_AX,Y_AX, mult=None):
  n = X.shape[0]
  scores = Y_AX(y,a,X )
  props = e_AX(a, X)
  if mult is not None:
    final = (( (A==a) *(Y==y)* (1.0 / props) - ((A==a)-props)*scores/props) )*mult
  else:
    final = (( (A==a) *(Y==y)* (1.0 / props) - ((A==a)-props)*scores/props) )#((A==a) * (1.0 / props)*((Y==y)-scores)) + scores
  return final
def IPW_score_singlearrival_vec_a(X,A,a,Y,y,e_AX,Y_AX, mult=None):
  props = e_AX(a, X)
  if mult is not None:
    final = (( (A==a) *(Y==y)* (1.0 / props)))*mult
  else:
    final =  (A==a) *(Y==y)* (1.0 / props)#((A==a) * (1.0 / props)*((Y==y)-scores)) + scores
  return final


def backwards_induction_evaluation_single_arrival(pi_e, config_, margSs,  # problem parameters
                                        X_, A_, Y_,
                                        G_s, R,  # problem mapping
                                        policy_score, e_AX, Y_AX,
                                        knn_x_policy=False, normalize=False):  # estimation strategy and nuisances
  '''
  Solve for policy by backwards induction

  '''

  dS = len(margSs)  # one dimensional representation
  nA = config_.nA
  [N, horizon, p] = X_.shape
  V = np.zeros([horizon,dS])

  even_t = range(horizon)[::2]
  odd_t = range(horizon)[1::2]
  X_even = X_[:,even_t,:].reshape(N*len(even_t), p); A_even = A_[:,even_t].reshape(N*len(even_t));Y_even = Y_[:,even_t].reshape(N*len(even_t))
  X_odd = X_[:, odd_t, :].reshape(N * len(odd_t), p);A_odd = A_[:, odd_t].reshape(N * len(odd_t));Y_odd = Y_[:, odd_t].reshape(N * len(odd_t))


  for t in reversed(range(horizon)):
    if t%2==0:
      X_parity = X_even; A_parity = A_even; Y_parity = Y_even; n_parity = len(even_t)
    else:
      X_parity = X_odd;A_parity = A_odd;Y_parity = Y_odd; n_parity = len(odd_t)
    for s in margSs:
        if s > 0:
          nS = len(G_s.map_[s]);policy_a = np.zeros([N*n_parity,nA])
          transitions_Y=np.zeros([len(G_s.map_[s]),nA])
          for sind,(sprime, y) in enumerate(G_s.map_[s]):
            transitions_Y_a=np.zeros(nA)

            for a in range(nA):
              # without sample splitting
              # policy_a[:,a] = pi_e(t,s,a,X_[:,t,:]) # off pol estimate of transition
              # transitions_Y_a[a] = np.mean(policy_score(X_[:, t, :], A_[:, t], a, Y_[:, t], y, e_AX, Y_AX,mult=policy_a[:,a]))
              policy_a[:, a] = pi_e(t, s, a, X_parity)
              transitions_Y_a[a] = np.mean(policy_score(X_parity, A_parity, a, Y_parity, y, e_AX, Y_AX, mult=policy_a[:, a]))

              transitions_Y[sind,a] = transitions_Y_a[a]
          # print('transition y', transitions_Y.sum())
          transitions_Y=transitions_Y/transitions_Y.sum()
          # print('after',transitions_Y)
          for sind, (sprime, y) in enumerate(G_s.map_[s]):
            if t < horizon - 1:
              V[t,s] += sum([ transitions_Y[sind,a]* (R(s, int(a), sprime) + V[t + 1, sprime]) for a in range(nA) ])
            else:
              V[t,s] += sum([ transitions_Y[sind,a] * R(s, int(a), sprime) for a in range(nA) ])

  return V


def delta_V(V,t,s,horizon,margSs):
  if t < horizon:
    if s > 0:
      delV = V[t,s-1] - V[t,s]
    else:
      delV = 0
  else:
    return 0
  return delV


def revenue_estimate(config_, pi_1, X, scores_, V, G_s, R, t,s, scores_or_oracle = True):
  '''
  if scores=True: scores is dictionary of (y,a) score vectors
  Y_AX is oracle outcome model
  '''
  horizon = config_.horizon
  if scores_or_oracle:
    if t < horizon - 1:
      return np.mean(sum([pi_1 *  scores_[(y, 1)]  * (R(s, 1, sprime) + V[t + 1, sprime]) \
                                    + (1 - pi_1) *  scores_[(y, 0)]  * (R(s, 0, sprime) + V[t + 1, sprime]) for (sprime, y) in
                                    G_s.map_[s]]))
    else:
      return np.mean(sum([pi_1 *  scores_[(y, 1)]  * (R(s, 1, sprime) ) \
                                    + (1 - pi_1) *  scores_[(y, 0)]  * (R(s, 0, sprime) ) for (sprime, y) in
                                    G_s.map_[s]]))
  else:
    Y_AX = scores_
    Y_11 = Y_AX(config_.action_price_map[1], X, 0, retprob=True)
    Y_10 = Y_AX(config_.action_price_map[0], X, 0, retprob=True)
    if t < horizon - 1:
      return np.mean(sum([pi_1 * (y*Y_11 + (1-y)*(1-Y_11)) * (R(s, 1, sprime)+ V[t + 1, sprime]) \
                        + (1 - pi_1) * (y*Y_10 + (1-y)*(1-Y_10)) * (R(s, 0, sprime)+ V[t + 1, sprime]) for (sprime, y) in
                        G_s.map_[s]]))
    else:
      return np.mean(sum([pi_1 * (y*Y_11 + (1-y)*(1-Y_11)) * (R(s, 1, sprime)) \
                        + (1 - pi_1) * (y*Y_10 + (1-y)*(1-Y_10)) * (R(s, 0, sprime)) for (sprime, y) in
                        G_s.map_[s]]))

def backwards_induction_polopt_threshold_search_single_arrival(config_, margSs,  # problem parameters
                                        X_, A_, Y_,
                                        G_s, R,  # problem mapping
                                        policy_score, e_AX, Y_AX_actionspace, XGenerator_parallel, Y_AX_oracle,
                                        knn_x_policy=False, normalize=False):  # estimation strategy and nuisances
  '''
  Solve for policy by backwards induction
  e_AX, Y_AX, are estimates
  '''

  dS = len(margSs)  # one dimensional representation
  nA = config_.nA
  [N, horizon, p] = X_.shape
  V_est = np.zeros([horizon,dS]); V = np.zeros([horizon,dS]); V_bias = np.zeros([horizon,dS])

  even_t = range(horizon)[::2]
  odd_t = range(horizon)[1::2]
  X_even = X_[:,even_t,:].reshape(N*len(even_t), p); A_even = A_[:,even_t].reshape(N*len(even_t));Y_even = Y_[:,even_t].reshape(N*len(even_t))
  X_odd = X_[:, odd_t, :].reshape(N * len(odd_t), p);A_odd = A_[:, odd_t].reshape(N * len(odd_t));Y_odd = Y_[:, odd_t].reshape(N * len(odd_t))
  nxoracle=1000
  X__o = XGenerator_parallel(nxoracle,horizon).reshape([nxoracle*horizon,p])

  policies = np.zeros([horizon, dS]) # store the thresholds
  policies_biased = np.zeros([horizon, dS])
  for t in reversed(range(horizon)):
    if t%2==0:
      X_parity = X_even; A_parity = A_even; Y_parity = Y_even; n_parity = len(even_t)
    else:
      X_parity = X_odd;A_parity = A_odd;Y_parity = Y_odd; n_parity = len(odd_t)
    for s in margSs:
        if s > 0:

          scores_ = dict() ; y_scores_ = dict();  y_scores_0=dict()
          for (sprime, y) in G_s.map_[s]:
            for a in range(nA):
              scores_[(y, a)] = policy_score(X_parity, A_parity, a, Y_parity, y, e_AX, Y_AX_actionspace)
              y_scores_[(y, a)] = Y_AX_actionspace(y, a, X_parity)
              y_scores_0[(y, a)] = Y_AX_actionspace(y, a, X__o)

          ratios = y_scores_[(1,1)] / y_scores_[(1,0)]
          n_thresh=200;estimates = np.zeros(n_thresh)
          thresholds = np.linspace( min(ratios), max(ratios), n_thresh )
          for t_ind, threshold in enumerate(thresholds):
            pi_1 = np.asarray(ratios > threshold).astype(int).flatten()
            estimates[t_ind] = revenue_estimate(config_,pi_1, X_parity, scores_, V_est, G_s, R, t,s, scores_or_oracle=True)
            # if t < horizon-1:
            #   estimates[t_ind] = np.mean(sum([pi_1 * scores_[(y, 1)] * (R(s, 1, sprime) + V[t + 1, sprime]) \
            #                 + (1 - pi_1) * scores_[(y, 0)] * (R(s, 0, sprime) + V[t + 1, sprime]) for (sprime, y) in
            #                 G_s.map_[s]]))
            # else:
            #   estimates[t_ind] = np.mean(sum([pi_1 * scores_[(y, 1)] * (R(s, 1, sprime)) \
            #                 + (1 - pi_1) * scores_[(y, 0)] * (R(s, 0, sprime)) for (sprime, y) in
            #                 G_s.map_[s]]))
          t_opt = np.argmax(estimates)
          # print('opt thresh', thresholds[t_opt])
          policies[(t,s)] = thresholds[t_opt]
          V_est[t, s] = estimates[t_opt]
          # estimate expected value out of sample
          pi_1 = np.asarray( y_scores_0[(1, 1)] / y_scores_0[(1, 0)] > thresholds[t_opt]).astype(int).flatten()

          V[t,s] = revenue_estimate(config_,pi_1, X__o, Y_AX_oracle, V, G_s, R, t,s, scores_or_oracle=False)

          # Estimate biased threshold
          deltaV_t1 = delta_V(V_bias, t+1, s,horizon,margSs)
          biased_threshold = (R(s,0,s-1) +deltaV_t1)/ (R(s,1,s-1) +deltaV_t1)
          pi_1 = np.asarray(ratios > biased_threshold).astype(int)
          V_bias[t,s] = revenue_estimate(config_,pi_1, X_parity, scores_, V, G_s, R, t,s, scores_or_oracle=True)
          policies_biased[(t,s)] = biased_threshold
  return [V, V_est, V_bias, policies, policies_biased]


def backwards_induction_with_estimates_single_arrival(config_, margSs,  # problem parameters
                                        X_, A_, Y_,
                                        G_s, R,  # problem mapping
                                        score, e_AX, Y_AX, Y_AX_prob_oracle,
                                        mu, sigma, nXoracle,
                                        knn_x_policy=False, normalize=False):  # estimation strategy and nuisances
  '''
  Solve for policy by backwards induction

  '''

  dS = len(margSs)  # one dimensional representation
  nA = config_.nA
  [N, horizon, p] = X_.shape
  price_action_map = config_.action_price_map
  V_exp_oracle = np.zeros([horizon,dS])
  V = np.zeros([horizon,dS])
  X = np.random.multivariate_normal(mu, sigma, size=nXoracle)
  knns_opt_pol = dict()
  for t in reversed(range(horizon)):
    for s in margSs:
        if s > 0:
          scores = np.zeros([N, nA])
          for a in range(nA):
            for (sprime, y) in G_s.map_[s]:
              if normalize:
                props = np.asarray([e_AX(a, X_[i,t, :]) for i in range(N)])
                # print( np.mean( (A_[:,t]==a)/props ))
                transitions_Y_a = score(X_[:, t, :], A_[:, t], a, Y_[:, t], y, e_AX, Y_AX)/np.mean( (A_[:,t]==a)/props )
              else:
                transitions_Y_a = score(X_[:, t, :], A_[:, t], a, Y_[:, t], y, e_AX, Y_AX)
              if t < horizon - 1:
                scores[:, a] += np.clip(transitions_Y_a,0,1) * (R(s, a, sprime) + V[t + 1, sprime])
              else:
                scores[:, a] += np.clip(transitions_Y_a,0,1) * R(s, a, sprime)
            # range over x
          best_a = np.argmax(scores, axis=1)
          V[t, s] += np.mean(scores[:, best_a])
          if knn_x_policy:
            knnx = FaissKNeighbors();knnx.fit(X_[:,t,:], best_a);knns_opt_pol[(t,s)] = knnx



    # expected value benchmark
          Y_1X = [Y_AX_prob_oracle(price_action_map[1], X[i,:], 0) for i in range(nXoracle)]
          Y_0X = [Y_AX_prob_oracle(price_action_map[0], X[i,:], 0) for i in range(nXoracle)]
          if t < horizon - 1:
            crit_ratio = (price_action_map[1] + V[t + 1, s - 1] - V[t + 1, s]) / (
                    price_action_map[0] + V[t + 1, s - 1] - V[t + 1, s])
            opt_pol = [1 if Y_1X[i] / Y_0X[i] > crit_ratio else 0 for i in range(nXoracle)]
            pol_val = np.mean([Y_1X[i] * (price_action_map[1] + V[t + 1, s - 1]) + (1 - Y_1X[i]) * V[t + 1, s] if opt_pol[i] == 1 else
                               Y_0X[i] * (price_action_map[0] + V[t + 1, s - 1]) + (1 - Y_1X[i]) * V[t + 1, s] for i in
                               range(nXoracle)])
          else:
            crit_ratio = (price_action_map[1] ) / (price_action_map[0] )
            opt_pol = [1 if Y_1X[i] / Y_0X[i] > crit_ratio else 0 for i in range(nXoracle)]
            pol_val = np.mean([Y_1X[i] * price_action_map[1] if opt_pol[i] == 1 else Y_0X[i] * price_action_map[0]  for i in range(nXoracle)])


          V_exp_oracle[t, s] = pol_val

  return [V,V_exp_oracle, knns_opt_pol]

def draw_trajectories(N, horizon, p, env, pi_b, config_, verbose=False):
    '''

    :param N, horizon, p: N is number of trajectories
    :param env: gym environment
    :param pi_b: behavior policy
    :param config_: config object
    :param verbose:
    :return:
    '''
    X__ = np.zeros([N, horizon, p]);S__ = np.zeros([N, horizon])  # capacity space
    A__ = np.zeros([N, horizon]);K__ = np.zeros([N, horizon]); R__ = np.zeros([N, horizon]);Y__ = np.zeros([N, horizon])
    for i in range(N):
      if (i % 200 == 0) and verbose: print(i)
      [S_, X_, A_, K_, R_, Y_] = sample_trajectory(env, pi_b, config_)
      S__[i, :] = S_.flatten();
      X__[i, :, :] = X_;
      A__[i, :] = A_;
      R__[i, :] = R_;
      Y__[i, :] = Y_
    return [S__, X__, A__, K__, R__, Y__]


def sample_trajectory(env, policy, config):
  # reset env
  env.reset()
  horizon = config.horizon
  J = config.J
  p = config.p
  S_ = np.zeros([horizon, config.pS])# structural state
  X_ = np.zeros([horizon, p]); K_ = np.zeros(horizon); R_ = np.zeros(horizon); Y_ = np.zeros(horizon)
  X_[0, :] = env.state[J:J + p]
  K_[0] = env.state[-1]
  S_[0,:] = env.state[:J]
  A_ = np.zeros(horizon)
  for t in range(horizon):
    # policy takes in X,k,S
    action = policy(X_[t,:], K_[t], S_[t,0:J], t)

    [S, R, epover, null] = env.step(action)

    A_[t] = action
    R_[t] = R
    Y_[t] = R > 0
    # Step returns x_t+1
    if t < horizon-1:
      S_[t + 1, :] = S[0:J]
      X_[t + 1, :] = S[J:J + p]
      K_[t + 1] = S[-1]

  return [S_, X_, A_, K_, R_, Y_]


def sample_trajectory_precomputed(env, policy, config, X_, K_,Y_a):
  '''
  precompute potential outcomes and X for speed.
  :param env:
  :param policy:
  :param config:
  :param X_:
  :param Y_a:
  :return:
  '''
  # reset env
  horizon = config.horizon
  J = config.J
  p = config.p
  S_ = np.zeros([horizon, config.pS])# structural state
  R_ = np.zeros(horizon); Y_ = np.zeros(horizon)
  K_[0] = env.state[-1]
  S_[0,:] = env.state[:J]
  A_ = np.zeros(horizon)
  for t in range(horizon):
    # policy takes in X,k,S
    k = int(K_[t])
    action = policy(X_[t,:], K_[t], S_[t,0:J], t)
    A_[t] = action
    Y = Y_a[t,action]
    Y_[t] = Y
    feasible = np.all(S_[t,0:J] >= env.M[k, :])
    R_[t] = feasible * Y * env._price_from_a(k, action)
    S_prime = S_[t,0:J] - feasible * Y * env.M[k, :]

    if t < horizon-1:
      S_[t + 1, :] = S_prime

  return [S_, A_, R_, Y_]

from copy import deepcopy
def draw_parallel_trajectories(N, horizon, p, env, pi_b, config_,
                               XGenerator_parallel, Y_AX, verbose=False):
  '''

  :param N, horizon, p: N is number of trajectories
  :param env: gym environment
  :param pi_b: behavior policy
  :param config_: config object
  :param verbose:
  :return:
  '''
  nA = env.nA
  X__ = XGenerator_parallel(N,horizon)
  Y__a = np.zeros([N,horizon,nA])
  # pregenerate potential outcomes
  for a in range(nA):
    X_r = deepcopy(X__).reshape([N*horizon,p])
    Y__a[:,:,a] = Y_AX(env.action_price_map[a],X_r,0,retprob=False).reshape([N,horizon])
  S__ = np.zeros([N, horizon])  # capacity space
  A__ = np.zeros([N, horizon]);
  K__ = np.zeros([N, horizon]);
  R__ = np.zeros([N, horizon]);
  Y__ = np.zeros([N, horizon])
  for i in range(N):
    if (i % 200 == 0) and verbose: print(i)
    [S_, A_, R_, Y_] = sample_trajectory_precomputed(env, pi_b, config_, X__[i,:,:], K__[i,:], Y__a[i,:,:])
    S__[i, :] = S_.flatten(); A__[i, :] = A_;R__[i, :] = R_;Y__[i, :] = Y_
  return [S__, X__, A__, K__, R__, Y__]

#
# def sample_parallel_trajectory(env, policy, config):
#   # reset env
#   env.reset()
#   horizon = config.horizon
#   J = config.J
#   p = config.p
#   S_ = np.zeros([horizon, config.pS])# structural state
#   X_ = np.zeros([horizon, p]); K_ = np.zeros(horizon); R_ = np.zeros(horizon); Y_ = np.zeros(horizon)
#   X_[0, :] = env.state[J:J + p]
#   K_[0] = env.state[-1]
#   S_[0,:] = env.state[:J]
#   A_ = np.zeros(horizon)
#   for t in range(horizon):
#     # policy takes in X,k,S
#     action = policy(X_[t,:], K_[t], S_[0:J], t)
#
#     [S, R, epover, null] = env.step(action)
#
#     A_[t] = action
#     R_[t] = R
#     Y_[t] = R > 0
#     # Step returns x_t+1
#     if t < horizon-1:
#       S_[t + 1, :] = S[0:J]
#       X_[t + 1, :] = S[J:J + p]
#       K_[t + 1] = S[-1]
#
#   return [S_, X_, A_, K_, R_, Y_]

def get_policy_evaluation(seed, scores, methods, scoreparams, ENNs, config_, env, pi_b, pi_b_eval, pi_e, Gs_, R, nS, calibrate=False,verbose=False,
                          K_folds = 2):
  '''
  run ope with different scoring functions
  wrapper function for OPE methods
  pi_b samples actions
  pi_b_eval returns prob of action

  '''
  np.random.seed(seed)
  print('seed',seed)
  nscores = len(scores)
  H = config_.horizon;nA=config_.nA; p = config_.p; exp_oos_opt_vals = np.zeros([len(ENNs), nscores])
  [S___, X___, A___, K___, R___, Y___] = draw_trajectories(ENNs[-1], H, p, env, pi_b, config_) # draw from prefix, once
  for ind, N in enumerate(ENNs):
    S__=S___[0:N,:];
    X__=X___[0:N,:,:];
    A__=A___[0:N,:];
    K__=K___[0:N,:];
    R__=R___[0:N,:];
    Y__=Y___[0:N,:]
    # [S__, X__, A__, K__, R__, Y__] = draw_trajectories(N, H, p, env, pi_b, config_)
    while len(np.unique(Y__))==1: # redraw data if only one class
      [S__, X__, A__, K__, R__, Y__] = draw_trajectories(N, H, p, env, pi_b, config_)
    if verbose:
      print(np.mean(A__, axis=0))  # treated probabilities
      print(np.mean(Y__, axis=0))  # treated probabilities
    # compute nuisances
    # [e_AX, e_AXs] = get_e_AX_singlearrival(config_, X__[:,:], A__[:,:]) # omit last timestep of selling horizon, boundary condition
    # [Y_AX_actionspace, Y_AXs] = get_Y_AX_singlearrival(config_, X__[:,:], A__[:,:], Y__[:,:],calibrate=calibrate)
    def pi_b_eval(t, s, a, x):
      return e_AX(a, x)
    if 'drl' in methods or 'drl-marginalw' in methods or 'fqe' in methods:
      Qsx = fitted_Q_evaluation(config_, pi_e, S__, X__, A__, Y__, R__)

    # dm eval+opt
    for scoreind, score in enumerate(scores):
      method = methods[scoreind]
      if method == 'drl':
        # estimate nuisances
        [etas, lambdas, ws] = dens_ratio_estimation(pi_e, pi_b_eval, S__, X__, A__)
        nuisances = [Qsx, etas, lambdas, ws]
        val = forwards_evaluation_single_arrival(pi_e, config_, S__, X__, A__, Y__, R__, nuisances)
        exp_oos_opt_vals[ind, scoreind] = val
        del nuisances

      elif method=='drl-marginalw':
        [etas, lambdas, ws] = dens_ratio_estimation(pi_e, pi_b_eval, S__, X__, A__,marginal_w=True)
        nuisances = [Qsx, etas, lambdas, ws]
        val = forwards_evaluation_single_arrival(pi_e, config_, S__, X__, A__, Y__, R__, nuisances,marginal_w=True)
        exp_oos_opt_vals[ind, scoreind] = val
        del nuisances

      elif method=='fqe':
        policy_a=[None]*nA; data=[None]*nA
        for a in range(nA):
          S0=slice_S(S__,0)
          policy_a[a] = pi_e(0,S0, a * np.ones(N).astype(int), X__[:, 0, :])
          data[a] = np.hstack([S0, X__[:, 0, :], a * np.ones([N, 1])])
        # E_\pi [Q_0 (S_0, X_0, a) ]
        Q0pi = sum([Qsx[0].predict(data[a]) * policy_a[a] for a in range(nA)]).flatten()
        exp_oos_opt_vals[ind,scoreind] = np.mean(Q0pi)

      elif method=='mis':
        [etas, lambdas, ws] = dens_ratio_estimation(pi_e, pi_b_eval, S__, X__, A__,marginal_w=True)
        nuisances = [Qsx, etas, lambdas, ws]
        val = forwards_evaluation_single_arrival_mis(pi_e, config_, S__, X__, A__, Y__, R__, nuisances,marginal_w=True)
        exp_oos_opt_vals[ind, scoreind] = val

      else:
        # Compute by computing a score
        # with sample splitting
        # evaluation by stateful off policy evaluation
        # override default estimation by passing a choice of nuisance estimator
        vals = np.zeros(K_folds)
        # Restrict attention to nonzero-state
        # Use KFold
        kf = KFold(n_splits=K_folds,random_state=seed,shuffle=True)
        for train_index, test_index in kf.split(A__): # train/test splits in the space of trajectory indices
        # # Datasets can be imbalanced if we split first without regard for state-based censoring
        #   # Train nuisances where S > 0
          nonzero_S = np.where(S__[train_index,:] > 0)
          X_nzs = X__[train_index,:,:][nonzero_S[0], nonzero_S[1], :];
          A_nzs = A__[train_index,:][nonzero_S[0], nonzero_S[1]]
          Y_nzs = Y__[train_index,:][nonzero_S[0], nonzero_S[1]]

          [e_AX, e_AXs] = get_e_AX_singlearrival(config_,X_nzs, A_nzs)  # omit last timestep of selling horizon, boundary condition
          def pi_b_eval(t, s, a, x):
            return e_AX(a, x)

          if scoreparams[scoreind] is not None:
            param=scoreparams[scoreind]
            if param['name']=='nonparametricy':
              [Y_AX_actionspace, Y_AXs] = get_Y_AX_singlearrival(config_,X_nzs, A_nzs, Y_nzs,regressor=param['nuisance'],calibrate=True)
          else:
            [Y_AX_actionspace, Y_AXs] = get_Y_AX_singlearrival(config_, X_nzs, A_nzs, Y_nzs)

          V = backwards_induction_evaluation_single_arrival(pi_e, config_, range(nS),  # problem parameters
                                                          X__[test_index,:,:], A__[test_index,:], Y__[test_index,:], Gs_,
                                                          R, score, e_AX, Y_AX_actionspace,normalize=True)
          exp_oos_opt_vals[ind, scoreind] += (1.0/K_folds) * V[0, -1] # average over k-folds

        # without sample splitting
        # [e_AX, e_AXs] = get_e_AX_singlearrival(config_, X__[:,:], A__[:,:]) # omit last timestep of selling horizon, boundary condition
        # [Y_AX_actionspace, Y_AXs] = get_Y_AX_singlearrival(config_, X__[:,:], A__[:,:], Y__[:,:],calibrate=calibrate)
        # def pi_b_eval(t, s, a, x):
        #   return e_AX(a, x)
        # if scoreparams[scoreind] is not None:
        #   param = scoreparams[scoreind]
        #   if param['name'] == 'nonparametricy':
        #     [Y_AX_actionspace, Y_AXs] = get_Y_AX_singlearrival(config_, X__, A__, Y__,
        #                                                        regressor=param['nuisance'], calibrate=True)
        #
        # V = backwards_induction_evaluation_single_arrival(pi_e, config_, range(nS),  # problem parameters
        #                                                   X__, A__, Y__, Gs_, R, score, e_AX, Y_AX_actionspace,
        #                                                   normalize=True)
        exp_oos_opt_vals[ind, scoreind] = V[0, -1]

  del [S__, X__, A__, K__, R__, Y__]
  return exp_oos_opt_vals


def get_off_policy_learning_backwards_induction(seed, scores, scoreparams, ENNs, config_, env, pi_b, Gs_, R, nS, nXoracle =10000,verbose=False):
  '''
  run ope with different scoring functions
  pi_b samples actions
  '''
  np.random.seed(seed)
  print('seed',seed)
  nscores = len(scores)
  H = config_.horizon;nA=config_.nA; p = config_.p; exp_oos_opt_vals = np.zeros([len(ENNs), nscores])
  ths_res = dict()
  [S___, X___, A___, K___, R___, Y___] = draw_trajectories(ENNs[-1], H, p, env, pi_b, config_) # draw from prefix, once
  for ind, N in enumerate(ENNs):
    print('N, ',N)
    S__=S___[0:N,:];
    X__=X___[0:N,:,:];
    A__=A___[0:N,:];
    K__=K___[0:N,:];
    R__=R___[0:N,:];
    Y__=Y___[0:N,:]
    # [S__, X__, A__, K__, R__, Y__] = draw_trajectories(N, H, p, env, pi_b, config_)
    while len(np.unique(Y__))==1: # redraw data if only one class
      [S__, X__, A__, K__, R__, Y__] = draw_trajectories(N, H, p, env, pi_b, config_)
    if verbose:
      print(np.mean(A__, axis=0))  # treated probabilities
      print(np.mean(Y__, axis=0))  # treated probabilities
    [e_AX, e_AXs] = get_e_AX_singlearrival(config_, X__, A__)
    [Y_AX_actionspace, Y_AXs] = get_Y_AX_singlearrival(config_, X__, A__, Y__)
    # dm eval+opt
    for scoreind, score in enumerate(scores):
      print('scoreind,',scoreind)
      opt_params = scoreparams[scoreind]['opt_params']
      opt_params['nuisances'] = [e_AX,Y_AX_actionspace]
      # optimize
      [ths, V] = backwards_induction_optimization_single_arrival(config_, range(nS),  # problem parameters
                 X__, A__, Y__, Gs_, R, score, e_AX, Y_AX_actionspace,opt_params,normalize=True)
      ths_res[(N,scoreind)] = ths;


      # Evaluate online
      # def pi_star_eval(x, k, s, t):
      #   if s > 0:
      #     th = ths[(t, int(s))];
      #     prob = opt_params['POL_PROB_1'](th, x)
      #     A = np.random.binomial(1, prob)
      #     return A
      #   else:
      #     return 0
      # print('evaluating,')
      # [S_, X_, A_, K_, R_, Y_] = draw_trajectories(nXoracle,
      #                                        config_.horizon, config_.p, env, pi_star_eval, config_)
      # exp_oos_opt_vals[ind,scoreind] = np.mean(R_.sum(axis=1))
      # del [S_, X_, A_, K_, R_, Y_]

  res=dict({'ths_res': ths_res,'exp_oos_opt_vals':exp_oos_opt_vals, 'seed': seed})
  pickle.dump(res, open('pol_opt_res-seed-' + str(seed)+ '-.p', 'wb'))

      # evaluate via rollouts
      # print(ths[(0, 1)])
      # def pi_star_eval(x, k, s, t):
      #   print('t,s',t,s)
      #   if s>0:
      #     print(ths[(t, int(s))])
      #     th = ths[(t, int(s))]
      #     prob = opt_params['POL_PROB_1'](th, x)
      #     A = np.random.binomial(1, prob)
      #     return A
      #   else: return 0
      # [S, X, A, K, R, Y] = draw_trajectories(nXoracle, H, p, env, pi_star_eval, config_)
      # exp_oos_opt_vals[scoreind] = R.mean()
      # del [S, X, A, K, R, Y]
  del [S___, X___, A___, K___, R___, Y___]

  return [ths_res, exp_oos_opt_vals]



def get_off_policy_learning_simple_search(seed, scores, ENNs, config_, env, pi_b, Gs_, R, nS,
                                          XGenerator_parallel,Y_AX_oracle, nXoracle =10000,verbose=False):
  '''
  run ope with different scoring functions
  pi_b samples actions
  '''
  np.random.seed(seed)
  print('seed',seed)
  nscores = len(scores)
  H = config_.horizon;nA=config_.nA; p = config_.p;

  policies_ = dict(); policies_biased_ = dict(); vals = dict()
  [S___, X___, A___, K___, R___, Y___] = draw_trajectories(ENNs[-1], H, p, env, pi_b, config_) # draw from prefix, once
  for ind, N in enumerate(ENNs):
    print('N, ',N)
    S__=S___[0:N,:];
    X__=X___[0:N,:,:];
    A__=A___[0:N,:];
    K__=K___[0:N,:];
    R__=R___[0:N,:],
    Y__=Y___[0:N,:]
    # [S__, X__, A__, K__, R__, Y__] = draw_trajectories(N, H, p, env, pi_b, config_)
    while len(np.unique(Y__))==1: # redraw data if only one class
      [S__, X__, A__, K__, R__, Y__] = draw_trajectories(N, H, p, env, pi_b, config_)
    if verbose:
      print(np.mean(A__, axis=0))  # treated probabilities
      print(np.mean(Y__, axis=0))  # treated probabilities
    [e_AX, e_AXs] = get_e_AX_singlearrival(config_, X__, A__)
    [Y_AX_actionspace, Y_AXs] = get_Y_AX_singlearrival(config_, X__, A__, Y__)
    # dm eval+opt
    for scoreind, score in enumerate(scores):
      print('scoreind,',scoreind)
      # optimize
      [V, V_est,  V_bias, policies, policies_biased] = backwards_induction_polopt_threshold_search_single_arrival(config_, range(nS),  # problem parameters
                 X__, A__, Y__, Gs_, R, score, e_AX, Y_AX_actionspace,XGenerator_parallel, Y_AX_oracle, normalize=True)
      vals[(N,scoreind)] = V
      policies_[(N,scoreind)] = policies
      policies_biased_[(N, scoreind)] = policies_biased

  del [S___, X___, A___, K___, R___, Y___]
  del [Y_AX_actionspace, Y_AXs]
  return [vals, V_est, V_bias, policies_, policies_biased_]



def single_stage_revenue_estimate(env, pi_1, X, Y_AX, G_s, R, s):
  Y_11 = Y_AX(env.action_price_map[1], X, 0, retprob=True)
  Y_10 = Y_AX(env.action_price_map[0], X, 0, retprob=True)
  return np.mean(sum([pi_1 * (y*Y_11 + (1-y)*Y_11) * (R(s, 1, sprime)) \
                                  + (1 - pi_1) * (y*Y_10 + (1-y)*Y_10) * (R(s, 0, sprime)) for (sprime, y) in
                                  G_s.map_[s]]))

def single_stage_revenue_check_simple_search(seed, scores, ENNs, config_, env, pi_b, Gs_, R, nS, XGenerator_parallel,Y_AX,nXoracle =10000,verbose=False,*args,**kwargs):
  '''
  run ope with different scoring functions
  pi_b samples actions
  '''
  np.random.seed(seed)
  print('seed',seed)
  nscores = len(scores)
  H = config_.horizon;nA=config_.nA; p = config_.p;

  policies_ = dict(); policies_biased_ = dict();V_est=dict();V=dict(); V_biased=dict()
  [S___, X___, A___, K___, R___, Y___] = draw_parallel_trajectories(ENNs[-1], H, p, env, pi_b, config_,XGenerator_parallel,Y_AX) # draw from prefix, once
  nxoracle=1000
  X__o = XGenerator_parallel(nxoracle,H).reshape([nxoracle*H,p])

  for ind, N in enumerate(ENNs):
    S__=S___[0:N,:];
    X__=X___[0:N,:,:];
    A__=A___[0:N,:];
    K__=K___[0:N,:];
    R__=R___[0:N,:],
    Y__=Y___[0:N,:]


    [e_AX, e_AXs] = get_e_AX_singlearrival(config_, X__, A__)
    [Y_AX_actionspace, Y_AXs] = get_Y_AX_singlearrival(config_, X__, A__, Y__)
    # dm eval+opt
    for scoreind, score in enumerate(scores):
      # optimize
      scores_=dict(); y_scores_=dict(); y_scores_0=dict()
      t=0; s=3
      for (sprime, y) in Gs_.map_[1]:
        for a in range(nA):
          scores_[(y, a)] = score(X__[:, t, :], A__[:, t], a, Y__[:, t], y, e_AX, Y_AX_actionspace)
          y_scores_[(y, a)] = Y_AX_actionspace(y, a, X__[:, t, :])
          y_scores_0[(y, a)] = Y_AX_actionspace(y, a, X__o)

      ratios = y_scores_[(1, 1)] / y_scores_[(1, 0)]
      n_thresh = 500;estimates = np.zeros(n_thresh);thresholds = np.linspace(min(ratios), max(ratios), n_thresh)

      for t_ind, threshold in enumerate(thresholds):
        pi_1 = np.asarray(ratios > threshold).astype(int).flatten()
        estimates[t_ind] = np.mean(sum([pi_1 * scores_[(y, 1)] * (R(s, 1, sprime)) \
                                      + (1 - pi_1) * scores_[(y, 0)] * (R(s, 0, sprime)) for (sprime, y) in
                                        Gs_.map_[s]]))
      t_opt = np.argmax(estimates)
      V_est[(N,scoreind)] = estimates[t_opt]
      pi_1 = np.asarray(y_scores_0[(1, 1)]/y_scores_0[(1, 0)] > thresholds[t_opt]).astype(int).flatten()
      V[(N,scoreind)] = single_stage_revenue_estimate(env, pi_1, X__o, Y_AX, Gs_, R, s)
      # Estimate biased threshold
      s=3
      biased_threshold = (R(s, 0, s - 1) ) / (R(s, 1, s - 1))
      pi_biased = np.asarray(y_scores_0[(1, 1)]/y_scores_0[(1, 0)] > biased_threshold).astype(int).flatten()
      V_biased[(N, scoreind)] = single_stage_revenue_estimate(env, pi_biased, X__o, Y_AX, Gs_, R, s)

      policies_[(N,scoreind)] = thresholds[t_opt]
      policies_biased_[(N, scoreind)] = biased_threshold

  del [S___, X___, A___, K___, R___, Y___]
  del X__o
  del ratios


  return [policies_, policies_biased_, V, V_est, V_biased]
def get_opt_params_update(opt_params,x,a,y,G_s, R, V, s,t,config_):
  opt_params['X_']=deepcopy(x); opt_params['Y_'] =deepcopy(y); opt_params['A_']=deepcopy(a)
  opt_params['sequential_info'] = [G_s, R, V, s,t]
  opt_params['config_']=config_
  return opt_params



def logistic_pol_asgn_a(theta, x, a):
    ''' Requires an intercept term
    '''
    n = x.shape[0]
    theta = theta.flatten()
    if (x[:,-1] == np.ones(n)).all():
        x = x
    else: # otherwise augment data
        x = np.hstack([x, np.ones([n,1])]);
    if len(theta) == 1:
        logit = np.multiply(x, theta).flatten()
    else:
        logit = np.dot(x, theta).flatten()
    LOGIT_TERM_POS = np.ones(n)*1.0 / ( np.ones(n) + np.exp( -logit ))
    return LOGIT_TERM_POS*a + (1-LOGIT_TERM_POS)*(1-a)


def backwards_induction_optimization_single_arrival(config_, margSs,  # problem parameters
                                        X_, A_, Y_,
                                        G_s, R,  # problem mapping
                                        policy_score, e_AX, Y_AX,
                                        opt_params,
                                        knn_x_policy=False, normalize=False):  # estimation strategy and nuisances
  '''
  Solve for policy by backwards induction
  '''

  dS = len(margSs)  # one dimensional representation
  nA = config_.nA
  [N, horizon, p] = X_.shape
  V = np.zeros([horizon+1,dS]) # input V into optimization without checking time
  POL_PROB_1 = opt_params['POL_PROB_1']
  POL_PROB_a = logistic_pol_asgn_a

  # opt_inds, eval_inds = train_test_split(range(N),test_size = 0.5)
  ths = dict()



  even_t = range(horizon)[::2]
  odd_t = range(horizon)[1::2]
  X_even = X_[:,even_t,:].reshape(N*len(even_t), p); A_even = A_[:,even_t].reshape(N*len(even_t));Y_even = Y_[:,even_t].reshape(N*len(even_t))
  X_odd = X_[:, odd_t, :].reshape(N * len(odd_t), p);A_odd = A_[:, odd_t].reshape(N * len(odd_t));Y_odd = Y_[:, odd_t].reshape(N * len(odd_t))

  for t in reversed(range(horizon)):
    # if (t%2==0): print('t',t)
    if t%2==0:
      X_parity = X_even; A_parity = A_even; Y_parity = Y_even; n_parity = len(even_t)
    else:
      X_parity = X_odd;A_parity = A_odd;Y_parity = Y_odd; n_parity = len(odd_t)
    for s in margSs:
        if s > 0:
          nS = len(G_s.map_[s]);
          opt_params = get_opt_params_update(opt_params,X_[:, t, :], A_[:, t], Y_[:, t],G_s, R, V, s,t,config_)
          [th, losses] = opt_w_restarts_score_refactored(**opt_params)
          ths[(t,s)] = th
          policy_a = np.zeros([N*n_parity,nA])
          transitions_Y = np.zeros([len(G_s.map_[s]), nA])
          # evaluate policy
          for sind,(sprime, y) in enumerate(G_s.map_[s]):
            transitions_Y_a=np.zeros(nA)
            for a in range(nA):
              policy_a[:,a] = POL_PROB_a(th, X_parity,a)
              # off pol estimate of transition

              transitions_Y_a[a] = np.mean(policy_score(X_parity, A_parity, a, Y_parity, y, e_AX, Y_AX, mult=policy_a[:, a]))
              transitions_Y[sind, a] = transitions_Y_a[a]
          transitions_Y = transitions_Y / transitions_Y.sum()
          for sind, (sprime, y) in enumerate(G_s.map_[s]):
            if t < horizon - 1:
              V[t,s] += sum([ transitions_Y[sind,a]* (R(s, int(a), sprime) + V[t + 1, sprime]) for a in range(nA) ])
            else:
              V[t,s] += sum([ transitions_Y[sind,a] * R(s, int(a), sprime) for a in range(nA) ])


          del opt_params['X_']
          del opt_params['A_']
          del opt_params['Y_']
  # print('done optimizing')
  return [ths,V]






def slice_S(S_,t):
  N=S_.shape[0]
  if len(S_.shape) > 2:
    S_t = S_[:, t, :]
  else:
    S_t = S_[:, t].reshape([N, 1])
  return S_t

def fitted_Q_evaluation(config_, policy, S_,X_,A_,Y_,R_):
  nA = config_.nA
  [N, horizon, p] = X_.shape
  regressor = RandomForestRegressor
  Qsx = dict()

  for t in reversed(range(horizon)):
    policy_a=[None]*nA
    # for a in range(nA):
    Qsx[t] = regressor()
    S_t = slice_S(S_,t)
    data = np.hstack([ S_t, X_[:, t, :], A_[:,t].reshape([N,1]) ])
    if t < horizon-1:
      data_next = [None]*nA
      S_t1=slice_S(S_,t+1)
      for a in range(nA):
        policy_a[a] = policy(t+1, S_t1, a*np.ones(N).astype(int), X_[:, t + 1, :])
        data_next[a] = np.hstack([S_t1, X_[:, t+1, :], a*np.ones([N,1])])
      # E_\pi [Q_t+1 (S_t+1, X_t+1, a) ]
      Q_tplus1_pi = sum([ Qsx[t+1].predict(data_next[a])*policy_a[a] for a in range(nA)  ]).flatten()
      Qsx[t].fit(data, R_[:,t] + Q_tplus1_pi )
    # generate Q predictions from previous
    # fit Q
    else:
      Qsx[t].fit(data, R_[:,t])
  return Qsx

def dens_ratio_estimation(pi_e, pi_b, S_,X_,A_, marginal_w=False):
  [N,horizon,p] = X_.shape
  regressor = RandomForestRegressor

  pi_e_=np.asarray([pi_e(t,  slice_S(S_,t), A_[:,t].astype(int).flatten(),X_[:,t,:]) for t in range(horizon)]).T ;
  pi_b_=np.asarray([pi_b(t,  slice_S(S_,t),A_[:,t].astype(int).flatten(),X_[:,t,:]) for t in range(horizon)]).T
  etas = pi_e_/pi_b_
  lambdas = np.cumprod(etas,axis=1)
  ws = dict()

  for t in range(horizon)[1:]: #only relevant for t > 0
    S_t = slice_S(S_,t)
    data = np.hstack([S_t, X_[:, t, :]])
    ws[t] = regressor()
    if marginal_w:
      ws[t].fit(S_t, lambdas[:, t - 1])
    else:
      ws[t].fit(data, lambdas[:,t-1])
  return [etas, lambdas, ws]

def forwards_evaluation_single_arrival(pi_e, config_,  # problem parameters
                                        S_, X_, A_, Y_, R_, nuisances,
                                       marginal_w=False):  # estimation strategy and nuisances
  '''
  Solve for policy by backwards induction

  '''
  nA = config_.nA
  [N, horizon, p] = X_.shape
  [Q_sx, etas, lambdas, ws] = nuisances
  q = [None] * nA; data_a = [None]*nA; mu = [None] * horizon
  val = 0
  for t in range(horizon):
    S_t = slice_S(S_,t)
    data = np.hstack([S_t, X_[:, t, :], A_[:, t].reshape([N, 1])])
    # density ratio (1 for t=0)
    if t == 0:
      mu[t] = np.ones(N)
    else:
      if marginal_w:
        w_t = ws[t].predict(S_t)  # drop last column
      else:
        w_t = ws[t].predict(data[:, :-1])  # drop last column
      mu[t] = etas[:,t]*w_t
    q_t_ai = Q_sx[t].predict(data)
    for a in range(nA):
      data[:,-1] = np.ones(N)*a # replace action
      q[a] = Q_sx[t].predict(data)
    q_t_pi = sum([ q[a]*pi_e(t,S_t,a*np.ones(N).astype(int), X_[:, t, :]) for a in range(nA)  ]).flatten()
    if t == 0:
      val += np.mean(mu[t]*(R_[:,t] - q_t_ai) + q_t_pi)
    else:
      val += np.mean(mu[t] * (R_[:, t] - q_t_ai) + mu[t-1]*q_t_pi)
  return val

def forwards_evaluation_single_arrival_mis(pi_e, config_,  # problem parameters
                                        S_, X_, A_, Y_, R_, nuisances,
                                       marginal_w=False):  # estimation strategy and nuisances
  '''
  Solve for policy by backwards induction

  '''
  nA = config_.nA
  [N, horizon, p] = X_.shape
  [Q_sx, etas, lambdas, ws] = nuisances
  q = [None] * nA; data_a = [None]*nA; mu = [None] * horizon
  val = 0
  for t in range(horizon):
    S_t = slice_S(S_,t)
    data = np.hstack([S_t, X_[:, t, :], A_[:, t].reshape([N, 1])])
    if t == 0:
      mu[t] = np.ones(N)
    else:
      if marginal_w:
        w_t = ws[t].predict(S_t)  # drop last column
      else:
        w_t = ws[t].predict(data[:, :-1])  # drop last column
      mu[t] = etas[:,t]*w_t

    val += np.mean(mu[t]*R_[:,t])

  return val

def eval_policy_parallel(ths_, POL_PROB_1, scores, ENNs, config_, env, nXoracle, fast = False, fast_params=None):
  '''
  :param ths_: indexed by (N, scoreind)
  :param POL_PROB_1:
  :param scores:
  :param ENNs:
  :param config_:
  :param env:
  :param nXoracle:
  :param fast: if true, use pre_computed generation
  :return:
  '''
  # ths_ is the dict with Ns,scores
  vals = np.zeros([len(ENNs), len(scores)])
  for n_ind, N in enumerate(ENNs):
    for scoreind in range(len(scores)):
      ths__ = ths_[(N, scoreind)]

      def pi_star_eval(x, k, s, t):
        if s > 0:
          th = ths__[(t, int(s))];
          prob = POL_PROB_1(th, x)
          A = np.random.binomial(1, prob)
          return A
        else:
          return 0

      if not fast:
        [S, X, A, K, R, Y] = draw_trajectories(nXoracle,
                                             config_.horizon, config_.p, env, pi_star_eval, config_)
      else:
        [S, X, A, K, R, Y] = draw_parallel_trajectories(nXoracle,
                                             config_.horizon, config_.p, env, pi_star_eval, config_,**fast_params)
      eval_val = np.mean(R.sum(axis=1))
      vals[n_ind, scoreind] = eval_val
  return vals

# ### Debug code
# rd = np.array([[0.65953346, 0.43524589, 0.70816539, -0.44480714, -0.14945422,
#                 -2.45490417]])
#
#
# def R_AX(price, x, prob=True):
#   X_ = np.hstack([x, price])
#
#   if prob:
#     return price * 5 + price * abs(x).sum() * 10
#   #         return prob*50
#   else:
#     Y = np.random.binomial(1, prob).item()
#   return Y
#
#
# nX = 1000
# t = 1
# V = np.zeros([config_.horizon, nS])
# s = 1
# price_action_map = config_.action_price_map
#
# X = np.random.multivariate_normal(mu, sigma, size=nX)
# Y_1X = [Y_AX_prob(price_action_map[1], X[i, :], 0) for i in range(nX)]
# Y_0X = [Y_AX_prob(price_action_map[0], X[i, :], 0) for i in range(nX)]
# # plt.hist(Y_1X,alpha=0.5,color='r')
# # plt.hist(Y_0X,alpha=0.5,color='b')
# plt.hist(
#   [Y_1X[i] * R_AX(price_action_map[1], X[i, :]) - Y_0X[i] * R_AX(price_action_map[0], X[i, :]) for i in range(nX)])
# val1 = [Y_1X[i] * (R_AX(price_action_map[1], X[i, :]) + V[t + 1, s - 1]) + (1 - Y_1X[i]) * V[t + 1, s] for i in
#         range(nX)]
# val0 = [Y_0X[i] * (R_AX(price_action_map[0], X[i, :]) + V[t + 1, s - 1]) + (1 - Y_0X[i]) * V[t + 1, s] for i in
#         range(nX)]
# opt_pol = [1 if val1[i] > val0[i] else 0 for i in range(nX)]
# pol_val = np.mean([val1[i] if val1[i] > val0[i] else val0[i] for i in range(nX)])
# print(pol_val)
# print(np.mean(opt_pol))
#
# y1 = np.mean(
#   [Y_1X[i] * (R_AX(price_action_map[1], X[i, :]) + V[t + 1, s - 1]) + (1 - Y_1X[i]) * V[t + 1, s] for i in range(nX)])
# y0 = np.mean(
#   [Y_0X[i] * (R_AX(price_action_map[0], X[i, :]) + V[t + 1, s - 1]) + (1 - Y_0X[i]) * V[t + 1, s] for i in range(nX)])
# print(y1 if y1 > y0 else y0, 'ate')
